/*
 * api_wind.h
 *
 * Warum werden alle moeglichen Api-Header eingebunden?
 *
 *   Von Window abgeleitete Klassen sollen sich nicht darum kuemmern muessen,
 *   alle Header der in ihnen verwendeten Komponenten jeweils fuer sich einzu-
 *   binden.
 *
 * class Window
 *
 *   Ein rechteckiges Fenster mit Titelleiste. Hiervon werden die einzelnen
 *   Dialoge abgeleitet. Enthaelt bereits die "Exit"-Variable, aber die
 *   einzelnen Rueckgabe-Abruffunktionen muessen die abgeleiteten Klassen
 *   selber bereitstellen.
 *
 * bool exit
 *
 *   Dies wird von der Instanz, die das Fensterobjekt verwaltet, abgefragt.
 *   Normalerweise wird das Fenster dann jedoch nicht beendet, sondern
 *   lediglich:
 *   1. die Fensterausgabe abgefragt (get_exit_with())
 *   2. exit wieder auf "false" gesetzt
 *   3. das Fenster unsichtbar gemacht (siehe visible).
 *
 * void draw_text(std::string& s, const int x, const int y)
 *
 *   Zeichnet einen Schattentext in etwas hellerer Farbe als der des Fensters
 *   linksbuendig an die entsprechende Stelle. Das ist einfach ein nicht
 *   anklickbarer Erklaerungstext. Diese Funktion wird sehr haeufig benutzt
 *   in vielen Fenstern, darum kann man sie zu einem entsprechend kuerzeren
 *   Aufruf verallgemeinern.
 *
 */

#pragma once

#include "button.h"
#include "button_b.h"
#include "button_c.h"
#include "button_e.h"
#include "button_k.h"
#include "button_s.h"
#include "button_t.h"
#include "element.h"
#include "frame.h"
#include "list_dir.h"
#include "list_fil.h"
#include "list_lev.h"
#include "number.h"
#include "texttype.h"

namespace Api {
class Window : public Element {

private:

    bool        exit;
    Torbit      torbit;

    bool        draw_title_required;
    int         x_drawn;
    int         y_drawn;

    std::string title;
    std::string subtitle;

    // Funktionen
    void prepare_torbit();

public:

    // Wird 20 gesetzt. Fenster mssen mindestens 30 y-lang sein.
    static const int title_bar_height;

    // Konstruktoren und Destruktor
             //     wohin, x, y, x_len, y_len, Fenstertitel
             Window(Torbit&, const int, const int,
                    const unsigned, const unsigned, const std::string&);
             Window(const Window&);
    virtual ~Window();

    inline       Torbit& get_torbit  ()       { return torbit;   }
    inline const Torbit& get_torbit  () const { return torbit;   }
    inline std::string&  get_title   ()       { return title;    }
    inline std::string&  get_subtitle()       { return subtitle; }
           void          set_title   (const std::string&);
           void          set_subtitle(const std::string&);

    inline bool get_exit() const        { return exit;    }
    inline void set_exit(bool b = true) { exit = b;       }

    // Funktionen, die jeweils die untergeordneten Elementfunktionen
    // mit den Fensterkoordinaten als Offset aufrufen
    // Nur wichtig fuer von Window abgeleitete Klassen, nicht fuer die
    // Instanz, die das abgeleitete Fenster kontrolliert
    inline bool is_mouse_over(const Element& e) const
                { return e.is_mouse_here(this->get_x(), this->get_y()); }

    inline bool button_clicked(Button& b)
                { return b.clicked(this->get_x(), this->get_y()); }

    // Die Abkuerzfunktion fuers Textzeichnen in den Draw-Routinen der Subkl.
    void draw_text         (const std::string&, const int, const int,
                            const int = color[COL_TEXT]);
    void draw_centered_text(const std::string&, const int, const int,
                            const int = color[COL_TEXT]);

           virtual void draw(); // Korrekt, verbessert Element::draw()
                   void draw_title();

protected:

           virtual void draw_self  () = 0;
    inline virtual void undraw_self() {}

};
// Ende von class Window

}
// Ende von namespace Api

