/*
 * editor.cpp
 *
 */

#include "../other/myalleg.h"
#include <fstream>

#include "editor.h"

#include "../graphic/glob_gfx.h"
#include "../graphic/gra_lib.h"
#include "../level/obj_lib.h"
#include "../other/hardware.h"
#include "../other/help.h"
#include "../other/log.h"

Editor::Editor()
:
    exit(false),
    timer_ticks_offset(Help::timer_ticks - 1), // - 1 prevents mouse flicker
    ticks(0),
    mouse_cursor_offset(15),
    mouse_cursor(GraLib::get(Globals::file_bitmap_mouse), *osd),

    level(Globals::user.single_last_dir + Globals::user.single_last_file),
    map(level.size_x,  level.size_y,
        level.start_x, level.start_y),
    map_frames(level.size_x,  level.size_y,
               level.start_x, level.start_y),

    object(Object::MAX),

    frame_start_x(-1),
    frame_start_y(-1),
    draw_required(true),
    browser_terrain_last_page(0)
{
    osd->clear_to_color(color[COL_PINK]);

    map.set_torus_x(level.torus_x);
    map.set_torus_y(level.torus_y);

    for (int type = Object::TERRAIN; type != Object::MAX; ++type)
     for (Level::PosIt itr =  level.pos[type].begin();
                       itr != level.pos[type].end(); ++itr)
     if (itr->ob && itr->ob->cb) {
        object[type].push_back(EdGraphic(map, itr->ob, itr->x, itr->y));
        EdGraphic& eg = *object[type].rbegin();
        eg.set_draw_info();
        if (itr->ob->type == Object::TERRAIN) {
            if      (itr->dark) eg.set_mode(Cutbit::DARK_EDITOR);
            else if (itr->noow) eg.set_mode(Cutbit::NOOW_EDITOR);
            eg                    .set_mirror  (itr->mirr);
            eg                    .set_rotation(itr->rot );
        }
        else if (itr->ob->type == Object::HATCH) {
            eg.set_x_frame(eg.get_x_frames() - 1);
            eg.set_rotation(itr->rot);
        }
        else if (itr->ob->type == Object::TRAP) {
            eg.set_x_frame(1);
        }
    }

    // Panel vorbereiten
    for (unsigned int i = 0; i < MAX; ++i) {
        panel.push_back(Api::BitmapButton(
         *osd, GraLib::get(Globals::file_bitmap_edit_panel)));
        panel[i].set_x(i/2*40);
        panel[i].set_y(420 + i%2*30);
        panel[i].set_x_frame(i);

        switch (i) {
        case FILE_NEW:      panel[i].set_hotkey(KEY_N);         break;
        case FILE_EXIT:     panel[i].set_hotkey(KEY_ESC);       break;
        case FILE_SAVE:     panel[i].set_hotkey();              break;
        case FILE_SAVE_AS:  panel[i].set_hotkey(KEY_F12);       break;
        case GRID_2:        panel[i].set_hotkey(KEY_2);         break; // KEY_1 deaktiviert
        case GRID_10:       panel[i].set_hotkey(KEY_3);         break; // alle 3 Buttons,
        case GRID_16:       panel[i].set_hotkey(KEY_4);         break; // anders kontroll.
        case SELECT_ALL:    panel[i].set_hotkey(KEY_A);         break;
        case SELECT_FRAME:  panel[i].set_hotkey();              break; // Shift, and.kontr.
        case SELECT_ADD:    panel[i].set_hotkey();              break; // CTRL, and. kontr.
        case SELECT_COPY:   panel[i].set_hotkey(KEY_C);         break;
        case SELECT_DELETE: panel[i].set_hotkey(KEY_DEL);       break;
        case SELECT_MINUS:  panel[i].set_hotkey(KEY_MINUS_PAD); break; // Total:
        case SELECT_PLUS:   panel[i].set_hotkey(KEY_PLUS_PAD);  break; // !hotkey, hold +/-
        case SELECT_FLIP:   panel[i].set_hotkey(KEY_F);         break;
        case SELECT_ROTATE: panel[i].set_hotkey(KEY_R);         break;
        case SELECT_DARK:   panel[i].set_hotkey(KEY_D);         break;
        case SELECT_NOOW:   panel[i].set_hotkey(KEY_O);         break;
        case VIEW_ZOOM:     panel[i].set_hotkey(KEY_Z);         break;
        case HELP:          panel[i].set_hotkey(KEY_H);         break;
        case MENU_SIZE:     panel[i].set_hotkey(KEY_F1);        break;
        case MENU_SCROLL:   panel[i].set_hotkey(KEY_F2);        break;
        case MENU_LEMMING:  panel[i].set_hotkey(KEY_F3);        break;
        case MENU_SKILL:    panel[i].set_hotkey(KEY_F4);        break;

        case ADD_TERRAIN:   panel[i].set_hotkey(KEY_F5);        break;
        case ADD_STEEL:     panel[i].set_hotkey(KEY_F6);        break;
        case ADD_HATCH:     panel[i].set_hotkey(KEY_F7);        break;
        case ADD_GOAL:      panel[i].set_hotkey(KEY_F8);        break;
        case ADD_DECO:      panel[i].set_hotkey(KEY_F9);        break;
        case ADD_HAZARD:    panel[i].set_hotkey(KEY_F10);       break;

        default:;
        }
    }

    // If an original level is loaded, use appropriate grid
    if (Help::string_get_extension(Globals::user.single_last_file)
     == Globals::ext_level_orig) panel[GRID_2].set_on();

    browser_save     = 0;
    browser_bitmap   = 0;
    window_size      = 0;
    window_scroll    = 0;
    window_variable  = 0;
    window_skill     = 0;
    box_unsaved_data = 0;
}

Editor::~Editor()
{
    selection.clear();

    if (browser_save)     delete browser_save;
    if (browser_bitmap)   delete browser_bitmap;
    if (window_size)      delete window_size;
    if (window_scroll)    delete window_scroll;
    if (window_variable)  delete window_variable;
    if (window_skill)     delete window_skill;
    if (box_unsaved_data) delete box_unsaved_data;
}



// ############################################################################
// ############################################################################
// ############################################################################



void Editor::calculate() {

    // Frameskipping koennte den Block umschliessen, aber es macht nichts,
    // wenn der Editor mit bestmoeglicher Geschwindigkeit laeuft
    ticks++;

    if (browser_save) {
        browser_save->calculate();
        if (browser_save->get_exit()) {
            if (browser_save->get_exit_with()) {
                Globals::user.single_last_dir  = browser_save->get_current_dir();
                Globals::user.single_last_file = browser_save->get_current_file();
                save_lists_and_level();
                // Dies ist wichtig, falls der Speichern-Dialog vom Code, der
                // die Exit-Box behandelt, aufgerufen wurde
                if (panel[FILE_EXIT].get_on()) exit = true;
            }
            delete browser_save;
            browser_save = 0;
            panel[FILE_NEW]    .set_off();
            panel[FILE_EXIT]   .set_off();
            panel[FILE_SAVE]   .set_off();
            panel[FILE_SAVE_AS].set_off();
        }
    }

    else if (browser_bitmap) {
        for (int p = ADD_TERRAIN; p != MAX; ++p)
         if (panel[p].get_on()) {
            calculate_bitmap_browser(
               p == ADD_TERRAIN ? Globals::user.editor_last_dir_terrain
             : p == ADD_STEEL   ? Globals::user.editor_last_dir_steel
             : p == ADD_HATCH   ? Globals::user.editor_last_dir_hatch
             : p == ADD_GOAL    ? Globals::user.editor_last_dir_goal
             : p == ADD_DECO    ? Globals::user.editor_last_dir_deco
             : p == ADD_HAZARD  ? Globals::user.editor_last_dir_hazard
             :                    Globals::user.editor_last_dir_terrain);
            if (!browser_bitmap) panel[p].set_off();
            break;
        }
    }

    else if (window_size) {
        window_size->calculate();
        if (window_size->get_exit()) {
            int x = window_size->get_exit_with_x_to_adjust();
            int y = window_size->get_exit_with_y_to_adjust();

            for  (int type = Object::TERRAIN; type != Object::MAX; ++type)
             for (GraIt i = object[type].begin(); i != object[type].end();++i){
                i->set_ground(map);
                if (x != 0) i->set_x(i->get_x() + x);
                if (y != 0) i->set_y(i->get_y() + y);
            }
            delete window_size;
            window_size = 0;
            panel[MENU_SIZE].set_off();
            draw_required = true; // if we only did this for x != 0 || y != 0,
                                  // altering only the right and bottom
                                  // margins would not trigger a redraw.
        }
    } else if (window_scroll) {
        window_scroll->calculate();
        if (window_scroll->get_exit()) {
            delete window_scroll;
            window_scroll = 0;
            panel[MENU_SCROLL].set_off();
        }
    } else if (window_variable) {
        window_variable->calculate();
        if (window_variable->get_exit()) {
            delete window_variable;
            window_variable = 0;
            panel[MENU_LEMMING].set_off();
        }
    } else if (window_skill) {
        window_skill->calculate();
        if (window_skill->get_exit()) {
            delete window_skill;
            window_skill = 0;
            panel[MENU_SKILL].set_off();
        }
    }

    else if (box_unsaved_data) {
        box_unsaved_data->calculate();
        unsigned i = box_unsaved_data->get_button_clicked();
        if (i == 1) {
            if (Globals::user.single_last_file.empty()) {
                browser_save = new Api::SaveBrowser(
                 Globals::dir_levels,
                 Globals::ext_level,
                 Globals::user.single_last_dir,
                 Globals::user.single_last_file,
                 Api::SaveBrowser::search_criterion_level,
                 Api::SaveBrowser::new_box_overwrite_level);
                browser_save->set_info_level_name(level.get_name());
                browser_save->set_info_file_name(Globals::user.single_last_dir
                 + Globals::user.single_last_file);
                delete box_unsaved_data;
                box_unsaved_data = 0;
                // Anhand des Status von panel[FILE_EXIT] erkennt der Code
                // oben zum Speichern-unter-Dialog, ob anschliessend beendet
                // werden muss.
            }
            else {
                save_lists_and_level();
                delete box_unsaved_data;
                box_unsaved_data = 0;
                if (panel[FILE_NEW].get_on()) delete_everything();
                else                          exit = true;
                panel[FILE_NEW].set_off();
                panel[FILE_EXIT].set_off();
            }
        }
        else if (i > 1) {
            delete box_unsaved_data;
            box_unsaved_data = 0;
            if (i == 2) {
                if (panel[FILE_NEW].get_on()) delete_everything();
                else                          exit = true;
            }
            panel[FILE_NEW].set_off();
            panel[FILE_EXIT].set_off();
        }
    }

    else calculate_self();
}


// Dies koennen verschiedene Browser sein
void Editor::calculate_bitmap_browser(std::string& save_last_dir)
{
    Api::BitmapBrowser*& br = browser_bitmap;
    br->calculate();
    if (br->get_exit()) {
        const Object* ob = br->get_exit_with();
        GraLi* li = ob ? &object[ob->type] : 0;

        if (li && ob->cb) {
            li->push_back(EdGraphic(map, ob));
            GraIt a = --li->end();

            // Neues Terrain-Objekt mittig unter den Mauszeiger legen
            a->set_x(map.get_mouse_x() - a->get_xl()/2);
            a->set_y(map.get_mouse_y() - a->get_yl()/2);
            a->set_draw_info();
            if (ob->type == Object::HATCH) a->set_x_frame(a->get_x_frames()-1);
            else if (ob->type == Object::TRAP) a->set_x_frame(1);

            // Korrektur, falls nicht sichtbar, weil Maus zu hoch auf Karte
            // mit sehr kleiner yl
            if (map.get_yl() < Level::view_yl
             && a->get_y() < 0) a->set_y(0);

            selection.clear();

            Selection s;
            s.o = a;
            s.l = li;
            selection.push_back(s);
            draw_required = true;

            // As the original graphics sets have been scaled to fit L++'s
            // gameplay, use twice the current grid if it's not an even number.
            // If it's an L2 graphic, use the 16 x 16 grid. (That is 8 x 8
            // pixels in L2, and that game works with a strict 16 x 8 grid.)
            if (br->get_current_dir().find(Globals::dir_bitmap_orig_l2)
             != std::string::npos) {
                panel[GRID_2 ].set_off();
                panel[GRID_10].set_off();
                panel[GRID_16].set_on();
            }
            else if (br->get_current_dir().find(Globals::dir_bitmap_orig)
             != std::string::npos) {
                if (!panel[GRID_2 ].get_on()
                 && !panel[GRID_10].get_on()
                 && !panel[GRID_16].get_on()) panel[GRID_2 ].set_on();
            }
            // We've already positioned the object under the mouse.
            // Move the object a few pixels so that it's aligned to the grid.
            const int grid = get_grid();
            a->set_x(a->get_x() + grid/2 - Help::mod(a->get_x()+grid/2,grid));
            a->set_y(a->get_y() + grid/2 - Help::mod(a->get_y()+grid/2,grid));
        }
        save_last_dir = br->get_current_dir();
        if (ob && ob->type == Object::TERRAIN && ob->subtype == 0) {
            browser_terrain_last_page = br->get_page();
        }
        delete (br);
        br = 0;
    }
}



// ############################################################################
// ############################################################################
// ############################################################################



void Editor::save_lists_and_level()
{
    save_lists();

    Level l          (Globals::user.single_last_dir
                    + Globals::user.single_last_file);
    if (level != l) {
        level.built = Date();
        // Always save in the L++ format, and also provide the correct
        // file extension
        if (Help::string_get_extension(Globals::user.single_last_file)
         != Globals::ext_level) {
            Help::string_remove_extension(Globals::user.single_last_file);
            Globals::user.single_last_file += Globals::ext_level;
        }
        level.save_to_file(Globals::user.single_last_dir
                         + Globals::user.single_last_file);
    }
}



void Editor::save_lists()
{
    for (int type = Object::TERRAIN; type != Object::MAX; ++type) {
        level.pos[type].clear();
        for (GraIt i = object[type].begin(); i != object[type].end(); ++i) {
            // Ausserhalb des sichtbaren Bereichs keine Objekte abspeichern
            if (i->get_x()               >= level.size_x
             || i->get_y()               >= level.size_y
             || i->get_x() + i->get_xl() <= 0
             || i->get_y() + i->get_yl() <= 0) continue;
            // Sonst abspeichern!
            Level::Pos temp;
            temp.ob   = i->get_object();
            temp.x    = i->get_x();
            temp.y    = i->get_y();
            temp.mirr = i->get_mirror();
            temp.rot  = i->get_rotation();
            temp.dark = i->get_mode() == Cutbit::DARK_EDITOR;
            temp.noow = i->get_mode() == Cutbit::NOOW_EDITOR;
            level.pos[type].push_back(temp);
        }
    }
}

