/*
 * gameplay/gameplay.cpp
 *
 */

#include "gameplay.h"

#include "../graphic/gra_lib.h"
#include "../other/globals.h"
#include "../other/hardware.h"
#include "../other/help.h"

// Statische Konstanten initialisieren
const int      Gameplay::updates_per_second(NETWORK_UPDATES_PER_SECOND);
const unsigned Gameplay::turbo_times_faster_than_fast (8);
const unsigned Gameplay::timer_ticks_for_update_fast
               (Help::timer_ticks_per_second/60);
const unsigned Gameplay::timer_ticks_for_update_normal
               (Help::timer_ticks_per_second/NETWORK_UPDATES_PER_SECOND);
const unsigned Gameplay::timer_ticks_for_update_slow
               (Help::timer_ticks_per_second/4*3);

const int Gameplay::mouse_cursor_offset(15);
const int Gameplay::mouse_max_lemming_distance_x(16);
const int Gameplay::mouse_max_lemming_distance_u(26);
const int Gameplay::mouse_max_lemming_distance_d(12);

const unsigned Gameplay::updates_lets_go   (updates_per_second);
const unsigned Gameplay::updates_hatch_open(updates_per_second * 2);

const int Gameplay::block_s = 14; // Blocker-Abstand zur Seite
const int Gameplay::block_m = 20; // zur Seite, wenn der andere ein Miner ist
const int Gameplay::block_u = 16; // nach oben
const int Gameplay::block_d =  8; // nach unten

// Standardkonstruktor
Gameplay::Gameplay(Replay* rep)
:
    exit                   (false),
    filename               (determine_filename(rep)),
    level                  (Network::get_started()
                            ? Network::get_level() : filename),
    player_local           (0), // will be assigned in prepare_players
    trlo                   (0), // same
    malo                   (0), // same
    local_ticks            (0),
    update_last_exiter     (0),
    timer_tick_last_update (Help::timer_ticks),
    timer_tick_last_F1     (0),
    timer_tick_last_F2     (0),
    timer_tick_last_F12    (0),
    timer_ticks_for_update_client(timer_ticks_for_update_normal),

    map                    (level.size_x,  level.size_y,
                            level.start_x, level.start_y),
    steel_mask             (level.size_x,  level.size_y),
    effect                 (map),

    mouse_cursor(GraLib::get(Globals::file_bitmap_mouse),       *osd),
    replay_sign (GraLib::get(Globals::file_bitmap_game_replay), *osd),

    console_bg             (*osd,  0, 0, LEMSCR_X),
    console_tt             (*osd,  0, 0, LEMSCR_X), // wird gleich noch geaend.
    stats                  (*osd),

    replaying              (rep? true : false),
    replay_recalculate_need(false),
    replay_recalculate_from(0),

    rate_displayed         (1),

    window_gameplay        (0),
    special                (Object::MAX)
{
    osd->clear_to_color(color[COL_PINK]);

    if (rep) replay = *rep;
    else     replay.set_level_filename(filename);

    prepare_players(rep);
    prepare_level();

    // Panels hinzufuegen
    // Es werden immer alle hinzugefuegt, aber je nach Netzwerktaetigkeit
    // nicht alle gemalt. Raten-Veraenderer werden immer gemalt, aber nicht
    // immer auf Anklicken geprueft, naemlich nicht bei einem Replay eines
    // Mehrspielerspiels.
    // X-Positioniert werden die Buttons durch position_panels() beim Zeichnen.
    for (unsigned i = 0; i < MAX; ++i) {
        if (i == PAUSE) {
            panel.push_back(Api::BitmapButton(*osd,
             GraLib::get(Globals::file_bitmap_game_pause), 0, 420));
        }
        else {
            panel.push_back(Api::BitmapButton(*osd,
             GraLib::get(Globals::file_bitmap_game_panel)));
            panel[i].set_x_frame(i);
            panel[i].set_y(i < PAUSE ? 420 +   i   % 2 * 30
                                     : 420 + (i+1) % 2 * 30);
        }
        if (i > RATE_PLUS && i < STATE_SAVE) panel[i].set_warm();
    }

    // For rate minus/plus, see the longer comment in gamepl_a.cpp.
    panel[PAUSE]      .set_hotkey(Globals::user.key_pause);
    panel[ZOOM]       .set_hotkey(Globals::user.key_zoom);
    panel[SPEED_SLOW] .set_hotkey(Globals::user.key_speed_slow);
    panel[SPEED_FAST] .set_hotkey(Globals::user.key_speed_fast);
    panel[SPEED_TURBO].set_hotkey(Globals::user.key_speed_turbo);
    panel[STATE_SAVE] .set_hotkey(Globals::user.key_state_save);
    panel[STATE_LOAD] .set_hotkey(Globals::user.key_state_load);
    panel[RESTART]    .set_hotkey(Globals::user.key_restart);
    panel[NUKE]       .set_hotkey(Globals::user.key_nuke);

    // Den Chat korrekt darstellen
    console_bg.set_undraw_color(color[COL_PINK]);
    console_tt.set_undraw_color(color[COL_API_D]);
    console_tt.set_invisible();
    console_tt.set_x(text_length(font_med, Globals::user.get_name().c_str()) + 10);
    console_tt.set_xl(LEMSCR_X - console_tt.get_x());
    console_tt.set_scroll();
}



Gameplay::~Gameplay()
{
    cs.trap  .clear();
    cs.tribes.clear();

    effect.clear_all_lists();
}



std::string Gameplay::determine_filename(Replay* rep)
{
    if (rep) {
        std::string s = "./replay/a.txt";
        return rep->get_level_filename();
    }
    else if (Network::get_started()) {
        return Globals::empty_string;
    }
    // Normaler Einzelspieler
    else {
        return Globals::user.single_last_dir + Globals::user.single_last_file;
    }
}



// ############################################################################
// ############################################################################
// ############################################################################



void Gameplay::prepare_players(Replay* rep)
{
    if (!rep) {
        if (!Network::get_started()) {
            replay.add_player(0, Lem::GARDEN, Globals::user.get_name());
            replay.set_player_local(0);
        }
        else {
            // The permutation is ignored here, it will be taken into account
            // when hatches and goals are distributed among the players.
            for (Network::PlDatCIt pitr = Network::get_player_data().begin();
             pitr != Network::get_player_data().end(); ++pitr) {
                if (!pitr->spec) {
                    replay.add_player(pitr->number,
                     static_cast <Lem::Style> (pitr->style), pitr->name);
                }
            }
            replay.set_player_local(Network::get_number());
            replay.set_permu       (Network::get_permu());
        }
    }

    // Done setting up the replay. Now set up the State cs.
    // Team formation is by Lem::Style. Determine how many teams we have.
    std::vector <bool> style_active(Lem::STYLE_MAX, false);
    for (Replay::PlayerCIt itr = replay.get_players().begin();
     itr != replay.get_players().end(); ++itr)
     if (style_active[itr->style] == false) {
        style_active[itr->style] = true;
        cs.tribes.push_back(Tribe());
        cs.tribes.back().style = itr->style;
    }
    replay.shorten_permu_to(cs.tribes.size());

    // Assign players to teams.
    for (Replay::PlayerCIt itr = replay.get_players().begin();
     itr != replay.get_players().end(); ++itr) {
        Tribe::It titr = cs.tribes.begin();
        while (titr != cs.tribes.end() && titr->style != itr->style)
         ++titr;
        if (titr != cs.tribes.end()) {
            titr->masters.push_back(Tribe::Master());
            titr->masters.back().number = itr->number;
            titr->masters.back().name   = itr->name;
            if (itr->number == replay.get_player_local()) {
                player_local = itr->number;
                trlo         = &*titr;
            }
        }
    }
    // Find local master. We cannot do this in the last for/if, because
    // there are still masters being added and STL containers may reallocate.
    for (std::list <Tribe::Master> ::iterator
     mitr = trlo->masters.begin(); mitr != trlo->masters.end(); ++mitr)
     if (mitr->number == player_local) {
        malo = &*mitr;
        break;
    }
}



void Gameplay::prepare_level()
{
    // Die Karte erschaffen
    map       .set_torus_xy(level.torus_x, level.torus_y);
    steel_mask.set_torus_xy(level.torus_x, level.torus_y);
    cs.land   .set_torus_xy(level.torus_x, level.torus_y);
    cs.land   .resize      (level.size_x,  level.size_y);
    level.draw_to           (cs.land, &steel_mask);
    Lemming::set_static_maps(cs.land, steel_mask, map);
    Lemming::set_effect_manager(effect);
    map.clear_to_color(color[COL_BLACK]);
    // save_bmp("test_land.bmp",  cs.land   .get_al_bitmap(), 0);
    // save_bmp("test_steel.bmp", steel_mask.get_al_bitmap(), 0);

    effect.set_trlo(trlo);
    cs.clock         = level.seconds * updates_per_second;
    cs.clock_running = (cs.tribes.size() == 1);
    // set_show_clock() wird stets aktuell in Gameplay::draw() aufgerufen

    // Interaktive Objekte aufstellen
    // In umgekehrter Reihenfolge erzeugen
    const Level& lv = level;

    for (int type = Object::TERRAIN; type != Object::MAX; ++type)
     for (Level::PosIt i = lv.pos[type].begin(); i != lv.pos[type].end();++i)
     switch (type) {
     case Object::HATCH:
         special[type].push_back(EdGraphic(map, i->ob, i->x, i->y));
         special[type].back().set_rotation(i->rot);
         break;
     case Object::GOAL:
        goal.push_back(Goal(map, i->ob, i->x, i->y));
        break;
     case Object::TRAP:
        cs.trap.push_back(EdGraphic(map, i->ob, i->x, i->y));
        break;
     case Object::FLING:
        cs.fling.push_back(EdGraphic(map, i->ob, i->x, i->y));
        break;
     case Object::DECO:
     case Object::WATER:
     case Object::ONEWAY:
     case Object::TRAMPOLINE:
        special[type].push_back(EdGraphic(map, i->ob, i->x, i->y));
        break;
     default:
        break;
    }

    // Nummer an Klappen ggf. nach unten korrigieren, nie nach oben
    // Irgendwas % 1 ist immer 0, also hoeren wir spaetestens bei einer
    // Klappe oder einem Ausgang auf
    std::vector <EdGraphic>& hatch = special[Object::HATCH];
    while (hatch.size()  % cs.tribes.size() != 0
     && cs.tribes.size() % hatch.size()     != 0) hatch.pop_back();
    while (goal.size()   % cs.tribes.size() != 0
     && cs.tribes.size() % goal.size()      != 0) goal.pop_back();

    // Spieler initialisieren
    // p.skill kommt unten nach, wenn wir das richtige Panel gewaehlt haben
    for (size_t i = 0; i < cs.tribes.size(); ++i) {
        size_t i_permuted = replay.get_permu()[i];
        Tribe& t = cs.tribes[i];
        // Style und Name sind schon gesetzt beim Spielererschaffen.
        // Couldn't have resized the vector earlier because the map must exist.
        t.lemming        . resize(lv.lemmings);
        t.hatch_next     = i_permuted % hatch.size();
        t.lemmings_hatch = lv.lemmings;
        t.rate_min       = lv.rate;
        t.rate           = lv.rate;
        for (unsigned a  = 0; a < 8; ++a) {
            t.skill[a].ac = lv.skill[a].ac;
            t.skill[a].nr = lv.skill[a].nr;
        }

        // Ausga(e)ng(e) zuweisen - dies ist etwas geklauter Code vom
        // Klappen zuweisen, der direkt hierunter steht
        unsigned g = i_permuted % goal.size();
        do {
            goal[g].add_tribe(&t);
            g += cs.tribes.size();
            g %= goal.size();
        } while (!goal[g].has_tribe(&t));

        // Auf dem Panel unten eintragen
        stats.add_tribe(t);
    }
    if (trlo) stats.set_tribe_local(*trlo);
    stats.set_initial(trlo->lemmings_hatch);
    if (cs.tribes.size() == 1) stats.set_required(lv.required);

    // Bildschirm-Start-Koordinaten festlegen fuer den lokalen Spieler, wenn
    // wir nicht alleine sind: In die Mitte aller eigenen Klappen
    if (cs.tribes.size() > 1) {
        unsigned sum_h = 0;
        int sum_x = 0;
        int sum_y = 0;
        int h = trlo->hatch_next;
        do {
            sum_x += hatch[h].get_x() + hatch[h].get_object()->trigger_x;
            sum_y += hatch[h].get_y() + hatch[h].get_object()->trigger_y;
            ++sum_h;
            // Die nchste Klappe kommt an die Reihe
            h += cs.tribes.size();
            h %= hatch.size();
        } while (h != trlo->hatch_next);

        map.set_screen_x(sum_x / sum_h -  LEMSCR_X       / 2);
        map.set_screen_y(sum_y / sum_h - (LEMSCR_Y - 60) / 3);
    }

    // Fhigkeits-Panels beschriften
    for (unsigned i = 0; i < 8; ++i) {
        skill.push_back(Api::SkillButton(*osd, (i+1) * 40, 420,
         trlo->skill[i].ac, trlo->skill[i].nr, trlo->style));
        skill[i].set_hotkey(Globals::user.key_skill[trlo->skill[i].ac]);
        skill[i].set_hot();
        // We will need to be careful when checking for clicked buttons,
        // as there can be more than one skill with the same hotkey in play.
        // This is done in Gameplay's calculcate code for these buttons.
    }
    // Einen auswaehlen: Das muss in einer separaten Schleife geschehen,
    // weil durch Neuallokation die Adressen in der ersten sich ggf. aendern.
    for (unsigned i = 0; i < skill.size(); ++i)
     if (skill[i].get_number() > 0) {
        skill[i].set_on();
        for  (Tribe::It  t = cs.tribes.begin();  t != cs.tribes.end();  ++t)
         for (Tribe::MIt m = t->masters.begin(); m != t->masters.end(); ++m) {
            m->skill_sel = i;
        }
        break;
    }
    rate_displayed = lv.rate;

    state_manager.save_zero(cs);
}



void Gameplay::save_result()
{
    if (!trlo || !malo) return;

    // Nur Einzelspielermodus speichern, und keine abgespielten Replays
    // anderer Leute als Erfolg vermerken. Man kann natuerlich schummeln
    // und kurz vor Replay-Ende eingreifen - aber schummeln kann man ohnehin
    // schon durch Bearbeiten der User-Datei.
    // Wandernde soll man nicht aufhalten. :-)
    if (cs.tribes.size()     == 1
     && trlo->lemmings_saved >= level.required
     && malo->name           == Globals::user.get_name())
    {
        Result res(level.built, trlo->lemmings_saved,
         trlo->skills_used, update_last_exiter);
        Globals::user.set_level_result(filename, res);
    }
}
