/*
 * lemming/basher.cpp
 *
 * The basher will bash away the lowest rectangle of most frames again during
 * the next frame. This is to prevent the horizontal lines when he's bashing
 * inside a digger's pit. These lines will be commented with "re-bash".
 *
 * special_x
 *
 *   Misst, wie weit ein Lemming hinuntergestiegen ist beim Bohren. Wenn das
 *   zu viel war, hoert er auf zu bohren. Siehe ausfuehrlichen Kommentar unten.
 *
 *   The basher will go up again if he's walked through a very small pit while
 *   bashing, to counter Horus' bug.
 *
 * special_y
 *
 *   This is 1 until the lemming has completed half of its first swing. While
 *   it's 1, steel doesn't cancel the basher.
 *
 */

#include "ac.h"

void assign_basher(Lemming& l)
{
    l.assign_default(Lem::BASHER);
    l.set_special_y(1);
}



void update_basher(Lemming& l, const UpdateArgs& ua)
{
    int steel_hit = 0;

    switch (l.get_frame() % 16) {
    case 0:
        // Frame 0: Die Erde hinter dem Lemming muss entfernt werden,
        // damit nicht nur Halbkreise spter entfernt werden!
        steel_hit += l.remove_rectangle( -2, -16,   1,   1);
        break;

    case 1:
        steel_hit += l.remove_rectangle(  2, -16,   5, -13);
        break;

    case 2:
        steel_hit += l.remove_rectangle(  6, -16,   7, -13);
        break;

    case 3:
        // Re-bash (see comment at top)
        steel_hit += l.remove_rectangle(  6, -16,   7, -13);

        steel_hit += l.remove_rectangle(  8, -15,   9, -13);
        steel_hit += l.remove_rectangle(  2, -12,   9, -11);
        break;

    case 4:
        // Re-bash
        steel_hit += l.remove_rectangle(  2, -12,   9, -11);

        steel_hit += l.remove_rectangle( 10, -14,  10, -13);
        steel_hit += l.remove_rectangle( 10, -12,  11, -11);
        steel_hit += l.remove_rectangle(  2, -10,  11,  -7);
        break;

    case 5:
        l.set_special_y(0);

        // Re-bash
        steel_hit += l.remove_rectangle( 10, -12,  11, -11);
        steel_hit += l.remove_rectangle(  2, -10,  11,  -7);

        steel_hit += l.remove_rectangle(  2,  -6,  11,  -3);
        steel_hit += l.remove_rectangle(  6,  -2,  10,  -1);
        break;

    case 6:
        // Re-bash
        steel_hit += l.remove_rectangle(  2,  -6,  11,  -3);
        steel_hit += l.remove_rectangle(  6,  -2,  10,  -1);

        steel_hit += l.remove_rectangle(  4,  -2,   5,  -1);
        steel_hit += l.remove_rectangle(  4,   0,   9,   0);
        steel_hit += l.remove_rectangle(  4,   1,   7,   1);
        break;

    case 7:
        steel_hit += l.remove_rectangle(  2,  -2,   3,   1);
        break;

    case 10:
        // Gibt es noch etwas zu Bohren?
        // Hier nehmen wir nicht alle Pixel am Fu
        // oder am Kopf des Lemmings mit.

        // Auch die vom Lemming am weitesten entfernten Doppelpixel
        // werden nicht berbrckt.

        // Alles wre dagegen 12, -16, 21, +1

        if (l.count_solid(12, -14, 19, -3) < 30)
        // Tills Bohr-Problem: Nochmal auf hauchduenne Waende pruefen
         for (int x = 12; x <= 21; ++x) {
            if (l.solid_wall_height(x) > 8) break;
            // Nicht abgebrochen?
            if (x == 21) {
                l.assign(Lem::WALKER);
                l.set_frame(8);
            }
        }
        if (l.get_ac() == Lem::BASHER) l.move_ahead();
        break;

    case 11:
    case 12:
    case 13:
    case 14:
        l.move_ahead();
        break;
    }

    // Wurde Stahl getroffen?
    // Die Object-Sachen knnen ruhig noch ausgefhrt werden.
    if (steel_hit && l.get_special_y() == 0) {
        l.turn();
        l.assign(Lem::WALKER);
        l.play_sound(ua, Sound::STEEL);
    }

    // Object misst, wie viele Pixel krzlich hinunter gestiegen wurden
    // Ab 9 Object-Punkten wird der Lemming zum Faller.
    // Die Pixel, in dem der Lemming steht, mssen auch hohl sein, denn
    // sonst fllt ein Basher im Zipfel seines Ganges, aber die Kollegen
    // nicht!
    while (!l.is_solid() && !l.is_solid(0, 1) && !l.is_solid(0, 0)
             && l.get_special_x() < 9) {
        l.move_down(1);
        l.set_special_x(l.get_special_x() + 2);
    }
    // Get up again to counter the Horus bug, but never go up further than
    // the basher started initially.
    while (l.get_special_x() > 0 && l.is_solid() && l.is_solid(0, 1)) {
        l.move_up(1);
        l.set_special_x(l.get_special_x() - 2);
        if (l.get_special_x() < 0) l.set_special_x(0);
    }

    // Object zu hoch? Dann Faller werden, sonst special_x senken.
    if (l.get_special_x() > 8) {
        l.move_up(1);
        l.assign(Lem::FALLER);
    }
    else if (l.get_special_x() > 0) l.set_special_x(l.get_special_x() - 1);

    // Wenn wirklich alles okay war, nchstes Basher-Frame setzen!
    if (l.get_ac() == Lem::BASHER) l.next_frame();
}
