/*
 * other/user.cpp
 *
 */

#include "user.h"
#include "language.h"
#include "io.h"
#include "myalleg.h" // KEY_*

#include "../lemming/lem_enum.h" // Lem::AC_MAX
#include "../network/net_t.h" // PlayerData::name_max_length

// Nullresultat zum Zurueckgeben von User::get_level_result(...)
Result Result::zero = Result();

Result::Result()
:
    built         (Globals::empty_string),
    lemmings_saved(0),
    skills_used   (0),
    updates_used  (0)
{
}

Result::Result(const Date& d, const int ls, const int su, const int uu)
:
    built         (d),
    lemmings_saved(ls),
    skills_used   (su),
    updates_used  (uu)
{
}

Result::~Result()
{
}



bool Result::operator == (const Result& r) const
{
    return built       == r.built
     && lemmings_saved == r.lemmings_saved
     && skills_used    == r.skills_used
     && updates_used   == r.updates_used;
}

bool Result::operator != (const Result& r) const
{
    return ! (*this == r);
}

bool Result::operator < (const Result& r) const
{
    // Wir muessen nicht nur updates_used > r.updates_used kontrollieren, um
    // festzustellen, ob r.updates_used besser ist, sondern r.updates_used
    // darf auch nicht null sein. Das ist der (im normalen Spiel unerreichbare)
    // Update-Wert eines standardkonstruierten Result-Objektes.
    return built      != r.built          ? built          < r.built
     : lemmings_saved != r.lemmings_saved ? lemmings_saved < r.lemmings_saved
     : skills_used    != r.skills_used    ? skills_used    > r.skills_used
     : updates_used   != r.updates_used   ? updates_used   > r.updates_used
                                            && r.updates_used > 0
     : false; // alle gleich
}

bool Result::operator > (const Result& r) const
{
    return *this != r && ! (*this < r);
}



// ############################################################################
// ############################################################################
// ############################################################################



User::User(const std::string& newname)
:
    language               (Language::NONE),
    option_group           (0),

    scroll_border          (true),
    scroll_right           (true),
    mouse_speed            (15),
    scroll_speed           (5),
    scroll_torus_x         (2),
    scroll_torus_y         (2),
    multiple_builders      (true),

    screen_scaling         (2),
    screen_border_colored  (true),
    screen_windowed        (false),
    arrows_replay          (true),
    arrows_network         (true),
    debris_amount          (2),
    debris_type            (1),

    sound_volume           (10),

    key_skill              (Lem::AC_MAX, 0),

    single_last_dir        (Globals::dir_levels_single),
    single_last_file       (Globals::empty_string),
    network_last_dir       (Globals::dir_levels_network),
    network_last_file      (Globals::empty_string),
    replay_last_dir        (Globals::dir_replay),
    replay_last_file       (Globals::empty_string),

    editor_last_dir_terrain(Globals::dir_bitmap),
    editor_last_dir_steel  (Globals::dir_bitmap),
    editor_last_dir_hatch  (Globals::dir_bitmap),
    editor_last_dir_goal   (Globals::dir_bitmap),
    editor_last_dir_deco   (Globals::dir_bitmap),
    editor_last_dir_hazard (Globals::dir_bitmap)
{
    if (!newname.empty()) load(newname);

    key_skill[Lem::WALKER]     = KEY_W;
    key_skill[Lem::BASHER]     = KEY_E;
    key_skill[Lem::BUILDER]    = KEY_R;
    key_skill[Lem::PLATFORMER] = KEY_T;
    key_skill[Lem::DIGGER]     = KEY_D;
    key_skill[Lem::MINER]      = KEY_G;
    key_skill[Lem::BLOCKER]    = KEY_V;
    key_skill[Lem::EXPLODER]   = KEY_B;
    key_skill[Lem::EXPLODER2]  = KEY_B;

    key_skill[Lem::CLIMBER]    = KEY_Q;
    key_skill[Lem::FLOATER]    = KEY_Q;
    key_skill[Lem::JUMPER]     = KEY_A;
    key_skill[Lem::ROPER]      = KEY_C;

    key_force_left             = KEY_S;
    key_force_right            = KEY_F;
    key_rate_minus             = KEY_1;
    key_rate_plus              = KEY_2;
    key_pause                  = KEY_SPACE;
    key_speed_slow             = KEY_3;
    key_speed_fast             = KEY_4;
    key_speed_turbo            = KEY_5;
    key_restart                = KEY_F1;
    key_state_load             = KEY_F2;
    key_state_save             = KEY_F5;
    key_zoom                   = KEY_Z;
    key_nuke                   = KEY_F12;
    key_chat                   = KEY_ENTER;
}



User::~User()
{
}



const Result& User::get_level_result(const std::string& filename) const
{
    std::map <std::string, Result>::const_iterator itr = result.find(filename);
    if (itr == result.end()) return Result::zero;
    else return itr->second;
}



void User::set_level_result(const std::string& filename, const Result& r)
{
    Result& saved_result = result[filename];
    if (saved_result < r) saved_result = r;
}



void User::load(const std::string& newname)
{
    result.clear();
    name = newname;

    if ((int) name.size() > PlayerData::name_max_length)
     name.resize(PlayerData::name_max_length);

    std::string filename = Globals::dir_data_user
                         + newname + Globals::ext_level;
    std::vector <IO::Line> lines;
    IO::fill_vector_from_file(lines, filename);

    for (IO::LineIt i = lines.begin(); i != lines.end(); ++i) switch(i->type) {

    case '$':
        if      (i->text1 == Globals::user_single_last_dir        ) single_last_dir        = i->text2;
        else if (i->text1 == Globals::user_single_last_file       ) single_last_file       = i->text2;
        else if (i->text1 == Globals::user_network_last_dir       ) network_last_dir       = i->text2;
        else if (i->text1 == Globals::user_network_last_file      ) network_last_file      = i->text2;
        else if (i->text1 == Globals::user_replay_last_dir        ) replay_last_dir        = i->text2;
        else if (i->text1 == Globals::user_replay_last_file       ) replay_last_file       = i->text2;

        else if (i->text1 == Globals::user_editor_last_dir_terrain) editor_last_dir_terrain= i->text2;
        else if (i->text1 == Globals::user_editor_last_dir_steel  ) editor_last_dir_steel  = i->text2;
        else if (i->text1 == Globals::user_editor_last_dir_hatch  ) editor_last_dir_hatch  = i->text2;
        else if (i->text1 == Globals::user_editor_last_dir_goal   ) editor_last_dir_goal   = i->text2;
        else if (i->text1 == Globals::user_editor_last_dir_deco   ) editor_last_dir_deco   = i->text2;
        else if (i->text1 == Globals::user_editor_last_dir_hazard ) editor_last_dir_hazard = i->text2;
        break;

    case '#':
        if      (i->text1 == Globals::user_language               ) language               = i->nr1;
        else if (i->text1 == Globals::user_option_group           ) option_group           = i->nr1;

        else if (i->text1 == Globals::user_mouse_speed            ) mouse_speed            = i->nr1;
        else if (i->text1 == Globals::user_scroll_speed           ) scroll_speed           = i->nr1;
        else if (i->text1 == Globals::user_scroll_border          ) scroll_border          = i->nr1;
        else if (i->text1 == Globals::user_scroll_right           ) scroll_right           = i->nr1;
        else if (i->text1 == Globals::user_scroll_torus_x         ) scroll_torus_x         = i->nr1;
        else if (i->text1 == Globals::user_scroll_torus_y         ) scroll_torus_y         = i->nr1;
        else if (i->text1 == Globals::user_multiple_builders      ) multiple_builders      = i->nr1;

        else if (i->text1 == Globals::user_screen_scaling         ) screen_scaling         = i->nr1;
        else if (i->text1 == Globals::user_screen_border_colored  ) screen_border_colored  = i->nr1;
        else if (i->text1 == Globals::user_screen_windowed        ) screen_windowed        = i->nr1;
        else if (i->text1 == Globals::user_arrows_replay          ) arrows_replay          = i->nr1;
        else if (i->text1 == Globals::user_arrows_network         ) arrows_network         = i->nr1;
        else if (i->text1 == Globals::user_debris_amount          ) debris_amount          = i->nr1;
        else if (i->text1 == Globals::user_debris_type            ) debris_type            = i->nr1;

        else if (i->text1 == Globals::user_sound_volume           ) sound_volume           = i->nr1;

        else if (i->text1 == Globals::user_key_force_left         ) key_force_left         = i->nr1;
        else if (i->text1 == Globals::user_key_force_right        ) key_force_right        = i->nr1;
        else if (i->text1 == Globals::user_key_rate_minus         ) key_rate_minus         = i->nr1;
        else if (i->text1 == Globals::user_key_rate_plus          ) key_rate_plus          = i->nr1;
        else if (i->text1 == Globals::user_key_pause              ) key_pause              = i->nr1;
        else if (i->text1 == Globals::user_key_speed_slow         ) key_speed_slow         = i->nr1;
        else if (i->text1 == Globals::user_key_speed_fast         ) key_speed_fast         = i->nr1;
        else if (i->text1 == Globals::user_key_speed_turbo        ) key_speed_turbo        = i->nr1;
        else if (i->text1 == Globals::user_key_restart            ) key_restart            = i->nr1;
        else if (i->text1 == Globals::user_key_state_load         ) key_state_load         = i->nr1;
        else if (i->text1 == Globals::user_key_state_save         ) key_state_save         = i->nr1;
        else if (i->text1 == Globals::user_key_zoom               ) key_zoom               = i->nr1;
        else if (i->text1 == Globals::user_key_nuke               ) key_nuke               = i->nr1;
        else if (i->text1 == Globals::user_key_chat               ) key_chat               = i->nr1;
        else {
            Lem::Ac ac = Lem::string_to_ac(i->text1);
            if (ac != Lem::AC_MAX) key_skill[ac] = i->nr1;
        }
        break;

    case '<': {
        Result result_read(i->text2, i->nr1, i->nr2, i->nr3);
        // Wir nutzen hier, dass ein standardkontruiertes Datenobjekt immer
        // kleiner ist als die geladenen Daten. map::operator[] fuegt ja ein
        // neues standardkonstruiertes Objekt ein, wenn noch kein Eintrag mit
        // dem uebergebenen Schluessel existiert.
        if (result[i->text1] < result_read) result[i->text1] = result_read;
        break; }

    default:
        break;

    }
}



void User::save() const
{
    std::string filename = Globals::dir_data_user
                         + name + Globals::ext_level;
    std::ofstream file(filename.c_str());

    file
     << IO::LineHash  (Globals::user_language,                language)
     << IO::LineHash  (Globals::user_option_group,            option_group)
     << std::endl

     << IO::LineHash  (Globals::user_mouse_speed,             mouse_speed)
     << IO::LineHash  (Globals::user_scroll_speed,            scroll_speed)
     << IO::LineHash  (Globals::user_scroll_border,           scroll_border)
     << IO::LineHash  (Globals::user_scroll_right,            scroll_right)
     << IO::LineHash  (Globals::user_scroll_torus_x,          scroll_torus_x)
     << IO::LineHash  (Globals::user_scroll_torus_y,          scroll_torus_y)
     << IO::LineHash  (Globals::user_multiple_builders,       multiple_builders)
     << std::endl

     << IO::LineHash  (Globals::user_screen_scaling,          screen_scaling)
     << IO::LineHash  (Globals::user_screen_border_colored,   screen_border_colored)
     << IO::LineHash  (Globals::user_screen_windowed,         screen_windowed)
     << IO::LineHash  (Globals::user_arrows_replay,           arrows_replay)
     << IO::LineHash  (Globals::user_arrows_network,          arrows_network)
     << IO::LineHash  (Globals::user_debris_amount,           debris_amount)
     << IO::LineHash  (Globals::user_debris_type,             debris_type)
     << std::endl

     << IO::LineHash  (Globals::user_sound_volume,            sound_volume)
     << std::endl

     << IO::LineDollar(Globals::user_single_last_dir,         single_last_dir)
     << IO::LineDollar(Globals::user_single_last_file,        single_last_file)
     << IO::LineDollar(Globals::user_network_last_dir,        network_last_dir)
     << IO::LineDollar(Globals::user_network_last_file,       network_last_file)
     << IO::LineDollar(Globals::user_replay_last_dir,         replay_last_dir)
     << IO::LineDollar(Globals::user_replay_last_file,        replay_last_file)
     << std::endl

     << IO::LineDollar(Globals::user_editor_last_dir_terrain, editor_last_dir_terrain)
     << IO::LineDollar(Globals::user_editor_last_dir_steel,   editor_last_dir_steel)
     << IO::LineDollar(Globals::user_editor_last_dir_hatch,   editor_last_dir_hatch)
     << IO::LineDollar(Globals::user_editor_last_dir_goal,    editor_last_dir_goal)
     << IO::LineDollar(Globals::user_editor_last_dir_deco,    editor_last_dir_deco)
     << IO::LineDollar(Globals::user_editor_last_dir_hazard,  editor_last_dir_hazard)
     << std::endl

     << IO::LineHash  (Globals::user_key_force_left,  key_force_left)
     << IO::LineHash  (Globals::user_key_force_right, key_force_right)
     << IO::LineHash  (Globals::user_key_rate_minus,  key_rate_minus)
     << IO::LineHash  (Globals::user_key_rate_plus,   key_rate_plus)
     << IO::LineHash  (Globals::user_key_pause,       key_pause)
     << IO::LineHash  (Globals::user_key_speed_slow,  key_speed_slow)
     << IO::LineHash  (Globals::user_key_speed_fast,  key_speed_fast)
     << IO::LineHash  (Globals::user_key_speed_turbo, key_speed_turbo)
     << IO::LineHash  (Globals::user_key_restart,     key_restart)
     << IO::LineHash  (Globals::user_key_state_load,  key_state_load)
     << IO::LineHash  (Globals::user_key_state_save,  key_state_save)
     << IO::LineHash  (Globals::user_key_zoom,        key_zoom)
     << IO::LineHash  (Globals::user_key_nuke,        key_nuke)
     << IO::LineHash  (Globals::user_key_chat,        key_chat);
    for (size_t i = 0; i < key_skill.size(); ++i)
     if (key_skill[i] != 0) file << IO::LineHash(
     Lem::ac_to_string(static_cast <Lem::Ac> (i)), key_skill[i]);

    if (!result.empty()) file << std::endl;

    for (std::map <std::string, Result> ::const_iterator itr = result.begin();
     itr != result.end(); ++itr) {
        std::string bstr;
        bstr << itr->second.built;
        file << IO::LineAngle(itr->first, itr->second.lemmings_saved,
                                          itr->second.skills_used,
                                          itr->second.updates_used, bstr);
    }

    file.close();
}
