/*
 * other/user.h
 *
 * Benutzerdaten, hauptsaechlich gespeicherte Levels
 *
 * struct LevelData enthaelt keinen Eintrag, der auf den Level selbst ver-
 * weist. Stattdessen wird pro Benutzer eine Hashmap benutzt, die einem
 * Dateinamen jeweils ein LevelResult-Objekt zuordnet.
 *
 * bool Result::operator < (const Result& r)
 *
 *   Liefert wahr, wenn *this ein schlechteres Ergebnis ist als r.
 *   Groessere Ergebnisse sind bessere Leistungen, oder aber built stimmt.
 *
 * void User::set_level_result(...)
 *
 *   Veraendert die Benutzerdaten des Levels, falls das LevelResult-Objekt
 *   ein besseres Ergebnis ist (Result::operator >). Besser ist das Ergebnis
 *   immer, wenn Level::built neuer ist als Result::built, bei Versions-
 *   gleichheit entscheidet die Zahl der geretteten Lemminge, bei erneuter
 *   Gleichheit die kuerzere Zeit.
 *
 */

#pragma once

#include <map>
#include <vector>

#include "date.h"

struct Result {
    Date built;
    int  lemmings_saved;
    int  skills_used;
    int  updates_used;

    static Result zero;

    Result ();
    Result (const Date&, const int, const int, const int);
    ~Result();

    bool    operator == (const Result&) const;
    bool    operator != (const Result&) const;
    bool    operator <  (const Result&) const;
    bool    operator >  (const Result&) const;
};



// ############################################################################
// ############################################################################
// ############################################################################



class User {

private:

    std::string name;

    std::map <std::string, Result> result;

public:

    ////////////
    // Optionen: Sind frei setzbar, wie damals in Globals
    ////////////

    int  language;
    int  option_group;

    bool scroll_border;
    bool scroll_right;
    int  mouse_speed;
    int  scroll_speed;
    int  scroll_torus_x;
    int  scroll_torus_y;
    bool multiple_builders;

    int  screen_scaling;
    bool screen_border_colored;
    bool screen_windowed;
    bool arrows_replay;
    bool arrows_network;
    int  debris_amount;
    int  debris_type;

    int  sound_volume;

    std::vector <int> key_skill;

    int  key_force_left;
    int  key_force_right;
    int  key_rate_minus;
    int  key_rate_plus;
    int  key_pause;
    int  key_speed_slow;
    int  key_speed_fast;
    int  key_speed_turbo;
    int  key_restart;
    int  key_state_load;
    int  key_state_save;
    int  key_zoom;
    int  key_nuke;
    int  key_chat;

    std::string single_last_dir;
    std::string single_last_file;
    std::string network_last_dir;
    std::string network_last_file;
    std::string replay_last_dir;
    std::string replay_last_file;

    std::string editor_last_dir_terrain;
    std::string editor_last_dir_steel;
    std::string editor_last_dir_hatch;
    std::string editor_last_dir_goal;
    std::string editor_last_dir_deco;
    std::string editor_last_dir_hazard;

    ////////////////
    // Funktionen //
    ////////////////

    User();
    User(const std::string&);
    ~User();

    inline const std::string& get_name() const { return name; }

    const Result& get_level_result(const std::string&) const;
    void          set_level_result(const std::string&, const Result&);

    void load(const std::string&);
    void save() const;

};
